"""Convert geoarrow.box arrays to geoarrow.polygon arrays."""

from __future__ import annotations

from typing import overload

import numpy as np
from arro3.core import (
    Array,
    ChunkedArray,
    Field,
    Schema,
    Table,
    fixed_size_list_array,
    list_array,
    struct_field,
)

from lonboard._geoarrow.extension_types import CoordinateDimension


@overload
def box_to_polygon(array: Array) -> Array: ...
@overload
def box_to_polygon(array: ChunkedArray) -> ChunkedArray: ...
def box_to_polygon(array: Array | ChunkedArray) -> Array | ChunkedArray:
    """Convert geoarrow.box arrays to geoarrow.polygon arrays."""
    if isinstance(array, ChunkedArray):
        return ChunkedArray([box_to_polygon(chunk) for chunk in array.chunks])

    if array.field.metadata_str["ARROW:extension:name"] != "geoarrow.box":
        raise ValueError(
            f"Expected geoarrow.box array in box_to_polygon, got {array.field.metadata_str['ARROW:extension:name']}",
        )

    if array.null_count > 0:
        raise ValueError("box_to_polygon does not currently support null values.")

    # Released arro3 doesn't have a way to get the children of the struct field
    # We check the number of child fields to infer the positions of the 2D bbox fields
    if array.type.num_fields == 4:
        xmin = struct_field(array, 0)
        ymin = struct_field(array, 1)
        xmax = struct_field(array, 2)
        ymax = struct_field(array, 3)
    elif array.type.num_fields == 6:
        xmin = struct_field(array, 0)
        ymin = struct_field(array, 1)
        xmax = struct_field(array, 3)
        ymax = struct_field(array, 4)
    elif array.type.num_fields == 8:
        xmin = struct_field(array, 0)
        ymin = struct_field(array, 1)
        xmax = struct_field(array, 4)
        ymax = struct_field(array, 5)
    else:
        raise ValueError(
            f"Unsupported number of fields in geoarrow.box array: {array.type.num_fields}",
        )

    # TODO: handle null values in the box array
    polygon_coords = np.zeros((len(array), 5, 2), dtype=np.float64)

    # https://github.com/geoarrow/geoarrow-rs/blob/c8561fb61bd21a4455a7b7741bf80a6935ae6060/rust/geoarrow-array/src/builder/geo_trait_wrappers.rs#L172-L181
    # ll, ul, ur, lr, ll
    polygon_coords[:, 0, 0] = xmin
    polygon_coords[:, 0, 1] = ymin

    polygon_coords[:, 1, 0] = xmin
    polygon_coords[:, 1, 1] = ymax

    polygon_coords[:, 2, 0] = xmax
    polygon_coords[:, 2, 1] = ymax

    polygon_coords[:, 3, 0] = xmax
    polygon_coords[:, 3, 1] = ymin

    polygon_coords[:, 4, 0] = xmin
    polygon_coords[:, 4, 1] = ymin

    ring_offsets = np.arange((len(array) + 1) * 5, step=5, dtype=np.int32)
    geom_offsets = np.arange(len(array) + 1, dtype=np.int32)

    arrow_coords = fixed_size_list_array(
        polygon_coords.ravel("C"),
        len(CoordinateDimension.XY),
    )
    arrow_rings = list_array(ring_offsets, arrow_coords)
    return list_array(geom_offsets, arrow_rings)


def parse_box_encoded_table(table: Table) -> Table:
    """Convert a table with a geoarrow.box column into a table with a geoarrow.polygon column."""
    # Find the index of the box column
    box_col_idx = None
    for idx in range(len(table.schema)):
        field = table.schema.field(idx)
        if field.metadata_str.get("ARROW:extension:name") == "geoarrow.box":
            if box_col_idx is not None:
                raise ValueError("Multiple geoarrow.box columns found in the table.")

            box_col_idx = idx

    if box_col_idx is None:
        # No box column found, return the table unchanged
        return table

    # Convert the box column to a polygon column
    box_col = table.column(box_col_idx)
    polygon_col = box_to_polygon(box_col)

    new_metadata = box_col.field.metadata_str.copy()
    new_metadata["ARROW:extension:name"] = "geoarrow.polygon"
    new_field = Field(
        name=box_col.field.name,
        type=polygon_col.type,
        nullable=box_col.field.nullable,
        metadata=new_metadata,
    )

    fields = list(iter(table.schema))
    fields[box_col_idx] = new_field
    new_schema = Schema(fields, metadata=table.schema.metadata)

    return table.set_column(
        box_col_idx,
        new_field,
        polygon_col,
    ).with_schema(new_schema)



"""High-level, super-simple API for visualizing GeoDataFrames."""

# ruff: noqa: C901, PLR0911, PLR0912, PLR0913, PLR0915

from __future__ import annotations

import json
import warnings
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Protocol, TypeAlias, cast

import numpy as np
from arro3.core import Array, ChunkedArray, Schema, Table, struct_field

from lonboard._compat import check_pandas_version
from lonboard._constants import EXTENSION_NAME
from lonboard._geoarrow.c_stream_import import import_arrow_c_stream
from lonboard._geoarrow.extension_types import construct_geometry_array
from lonboard._geoarrow.geopandas_interop import geopandas_to_geoarrow
from lonboard._geoarrow.parse_wkb import parse_serialized_table
from lonboard._geoarrow.row_index import add_positional_row_index
from lonboard._layer import PathLayer, PolygonLayer, ScatterplotLayer
from lonboard._map import Map
from lonboard._utils import (
    get_geometry_column_index,
    split_mixed_gdf,
    split_mixed_shapely_array,
)
from lonboard.basemap import CartoBasemap

if TYPE_CHECKING:
    import duckdb
    import geopandas as gpd
    import pyarrow as pa
    import shapely.geometry
    import shapely.geometry.base
    from arro3.core.types import ArrowArrayExportable, ArrowStreamExportable
    from numpy.typing import NDArray

    from lonboard.types.layer import (
        PathLayerKwargs,
        PolygonLayerKwargs,
        ScatterplotLayerKwargs,
    )
    from lonboard.types.map import MapKwargs

    class GeoInterfaceProtocol(Protocol):
        @property
        def __geo_interface__(self) -> dict: ...

    VizDataInput: TypeAlias = (
        gpd.GeoDataFrame
        | gpd.GeoSeries
        | pa.Table
        | NDArray[np.object_]
        | shapely.geometry.base.BaseGeometry
        | ArrowArrayExportable
        | ArrowStreamExportable
        | GeoInterfaceProtocol
        | dict[str, Any]
        | duckdb.DuckDBPyRelation
    )
    """A type definition for allowed data inputs to the `viz` function."""


# From mbview
# https://github.com/mapbox/mbview/blob/e64bd86cfe4a63e6af4ea1d310bd49be4f162a43/views/vector.ejs#L75-L87
COLORS = [
    "#FC49A3",  # pink
    "#CC66FF",  # purple-ish
    "#66CCFF",  # sky blue
    "#66FFCC",  # teal
    "#00FF00",  # lime green
    "#FFCC66",  # light orange
    "#FF6666",  # salmon
    "#FF0000",  # red
    "#FF8000",  # orange
    "#FFFF66",  # yellow
    "#00FFFF",  # turquoise
]
COLOR_COUNTER = 0
DEFAULT_POLYGON_LINE_COLOR = [0, 0, 0, 200]


def viz(
    data: VizDataInput | list[VizDataInput] | tuple[VizDataInput, ...],
    *,
    scatterplot_kwargs: ScatterplotLayerKwargs | None = None,
    path_kwargs: PathLayerKwargs | None = None,
    polygon_kwargs: PolygonLayerKwargs | None = None,
    map_kwargs: MapKwargs | None = None,
    con: duckdb.DuckDBPyConnection | None = None,
) -> Map:
    """Plot your data easily.

    The goal of this function is to make it simple to get _something_ showing on a map.
    For more control over rendering, construct `Map` and `Layer` objects directly.

    This function accepts a variety of geospatial inputs:

    - GeoPandas `GeoDataFrame`.
    - GeoPandas `GeoSeries`.
    - numpy array of Shapely objects.
    - Single Shapely object.
    - A DuckDB query with a spatial column from DuckDB Spatial.

        !!! warning

            The DuckDB query must be run with
            [`duckdb.sql()`](https://duckdb.org/docs/api/python/reference/#duckdb.sql)
            or
            [`duckdb.DuckDBPyConnection.sql()`](https://duckdb.org/docs/api/python/reference/#duckdb.DuckDBPyConnection.sql)
            and not with `duckdb.execute()` or `duckdb.DuckDBPyConnection.execute()`.

            For example

            ```py
            import duckdb
            from lonboard import viz

            sql = "SELECT * FROM spatial_table;"
            query = duckdb.sql(sql)
            viz(query)
            ```

            You can also render an entire table by using the `table()` method:

            ```py
            import duckdb
            from lonboard import viz

            con = duckdb.connect()
            con.execute("CREATE TABLE spatial_table AS ...;")
            viz(con.table("spatial_table"))
            ```

        !!! warning

            DuckDB Spatial does not currently expose coordinate reference system
            information, so the user must ensure that data has been reprojected to
            EPSG:4326.

    - Any Python class with a `__geo_interface__` property conforming to the
        [Geo Interface protocol](https://gist.github.com/sgillies/2217756).
    - `dict` holding GeoJSON-like data.
    - pyarrow `Table` with a geometry column marked with a
        [GeoArrow](https://geoarrow.org/) extension type.
    - pyarrow `Array` or `ChunkedArray` marked with a [GeoArrow extension type defined by geoarrow-pyarrow](https://geoarrow.org/geoarrow-python/main/pyarrow.html#geoarrow.pyarrow.GeometryExtensionType).
    - Arrow-compatible Array, ChunkedArray, Table, or RecordBatch objects that have associated GeoArrow metadata. An object is "Arrow-compatible" if it implements the [Arrow PyCapsule Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html) and has either an `__arrow_c_array__` or `__arrow_c_stream__` method. The provided Arrow data must be or have a geometry column marked with a GeoArrow extension type.

        Some examples of these sources include pyogrio's [`open_arrow`][pyogrio.open_arrow], DuckDB Spatial, [GeoArrow-Rust's Python bindings](https://geoarrow.org/geoarrow-rs/python/latest/), [GeoDataFusion database connections](https://github.com/datafusion-contrib/datafusion-geo), and, soon, [GeoPolars DataFrames](https://github.com/geopolars/geopolars).

    Alternatively, you can pass a `list` or `tuple` of any of the above inputs.

    If you want to easily add more data, to an existing map, you can pass the output of
    `viz` into [`Map.add_layer`][lonboard.Map.add_layer].

    Args:
        data: a data object of any supported type.

    Keyword Args:
        scatterplot_kwargs: a `dict` of parameters to pass down to all generated
            [`ScatterplotLayer`][lonboard.ScatterplotLayer]s.
        path_kwargs: a `dict` of parameters to pass down to all generated
            [`PathLayer`][lonboard.PathLayer]s.
        polygon_kwargs: a `dict` of parameters to pass down to all generated
            [`PolygonLayer`][lonboard.PolygonLayer]s.
        map_kwargs: a `dict` of parameters to pass down to the generated
            [`Map`][lonboard.Map].
        con: Deprecated: the active DuckDB connection. This argument has no effect and
            might be removed in the future.

    For more control over rendering, construct [`Map`][lonboard.Map] and `Layer` objects
    directly.

    Returns:
        widget visualizing the provided data.

    """
    global COLOR_COUNTER  # noqa: PLW0603 Using the global statement to update `COLOR_COUNTER` is discouraged

    if con is not None:
        warnings.warn(
            "The 'con' argument is deprecated and may be removed in a future version. "
            "It has no effect.",
            category=DeprecationWarning,
            stacklevel=2,
        )

    if isinstance(data, (list, tuple)):
        layers: list[ScatterplotLayer | PathLayer | PolygonLayer] = []
        for i, item in enumerate(data):
            ls = create_layers_from_data_input(
                item,
                _viz_color=COLORS[(COLOR_COUNTER + i) % len(COLORS)],
                scatterplot_kwargs=scatterplot_kwargs,
                path_kwargs=path_kwargs,
                polygon_kwargs=polygon_kwargs,
            )
            layers.extend(ls)

        COLOR_COUNTER += len(layers)
    else:
        layers = create_layers_from_data_input(
            data,
            _viz_color=COLORS[COLOR_COUNTER % len(COLORS)],
            scatterplot_kwargs=scatterplot_kwargs,
            path_kwargs=path_kwargs,
            polygon_kwargs=polygon_kwargs,
        )
        COLOR_COUNTER += 1

    map_kwargs = map_kwargs if map_kwargs else {}

    if "basemap_style" not in map_kwargs:
        map_kwargs["basemap_style"] = CartoBasemap.DarkMatter

    return Map(layers=layers, **map_kwargs)


DUCKDB_PY_CONN_ERROR = dedent("""\
    Must pass in DuckDBPyRelation object, not DuckDBPyConnection.

    Instead of using `duckdb.execute()` or `con.execute()`, use `duckdb.sql()`,
    `con.sql()` or `con.table()`.
    """)


def create_layers_from_data_input(
    data: VizDataInput,
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    """Create one or more renderable layers from data input.

    This helper function can create multiple layers in the case of mixed input.
    """
    # geopandas GeoDataFrame
    if (
        data.__class__.__module__.startswith("geopandas")
        and data.__class__.__name__ == "GeoDataFrame"
    ):
        return _viz_geopandas_geodataframe(data, **kwargs)  # type: ignore

    # geopandas GeoSeries
    if (
        data.__class__.__module__.startswith("geopandas")
        and data.__class__.__name__ == "GeoSeries"
    ):
        return _viz_geopandas_geoseries(data, **kwargs)  # type: ignore

    # duckdb DuckDBPyRelation
    if (
        data.__class__.__module__.startswith("duckdb")
        and data.__class__.__name__ == "DuckDBPyRelation"
    ):
        return _viz_duckdb_relation(data, **kwargs)  # type: ignore

    if (
        data.__class__.__module__.startswith("duckdb")
        and data.__class__.__name__ == "DuckDBPyConnection"
    ):
        raise TypeError(DUCKDB_PY_CONN_ERROR)

    # Shapely array
    if isinstance(data, np.ndarray) and np.issubdtype(data.dtype, np.object_):
        return _viz_shapely_array(data, **kwargs)

    # Shapely scalar
    if data.__class__.__module__.startswith("shapely") and any(
        cls.__name__ == "BaseGeometry" for cls in data.__class__.__mro__
    ):
        return _viz_shapely_scalar(data, **kwargs)  # type: ignore

    # Anything with __arrow_c_array__
    if hasattr(data, "__arrow_c_array__"):
        data = cast("ArrowArrayExportable", data)
        array = Array.from_arrow(data)
        ca = ChunkedArray([array])
        return _viz_geoarrow_chunked_array(ca, **kwargs)

    # Anything with __arrow_c_stream__
    if hasattr(data, "__arrow_c_stream__"):
        data = cast("ArrowStreamExportable", data)
        imported_stream = import_arrow_c_stream(data)
        if isinstance(imported_stream, Table):
            return _viz_geoarrow_table(imported_stream, **kwargs)

        assert isinstance(imported_stream, ChunkedArray)
        return _viz_geoarrow_chunked_array(imported_stream, **kwargs)

    # Anything with __geo_interface__
    if hasattr(data, "__geo_interface__"):
        data = cast("GeoInterfaceProtocol", data)
        return _viz_geo_interface(data.__geo_interface__, **kwargs)

    # GeoJSON dict
    if isinstance(data, dict):
        if data.get("type") in [
            "Point",
            "LineString",
            "Polygon",
            "MultiPoint",
            "MultiLineString",
            "MultiPolygon",
            "GeometryCollection",
            "Feature",
            "FeatureCollection",
        ]:
            return _viz_geo_interface(data, **kwargs)

        raise ValueError(
            "If passing a dict, must be a GeoJSON "
            "Geometry, Feature, or FeatureCollection.",
        )

    raise ValueError


def _viz_geopandas_geodataframe(
    data: gpd.GeoDataFrame,
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    layers: list[ScatterplotLayer | PathLayer | PolygonLayer] = []
    for partial_gdf in split_mixed_gdf(data):
        table = geopandas_to_geoarrow(partial_gdf)
        layers.extend(_viz_geoarrow_table(table, **kwargs))

    return layers


def _viz_geopandas_geoseries(
    data: gpd.GeoSeries,
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    import geopandas as gpd

    gdf = gpd.GeoDataFrame(geometry=data)  # type: ignore
    return _viz_geopandas_geodataframe(gdf, **kwargs)


def _viz_duckdb_relation(
    data: duckdb.DuckDBPyRelation,
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    from lonboard._geoarrow._duckdb import from_duckdb

    table = from_duckdb(data)
    return _viz_geoarrow_table(table, **kwargs)


def _viz_shapely_scalar(
    data: shapely.geometry.base.BaseGeometry,
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    return _viz_shapely_array(np.array([data]), **kwargs)


def _viz_shapely_array(
    data: NDArray[np.object_],
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    layers: list[ScatterplotLayer | PathLayer | PolygonLayer] = []
    for partial_geometry_array in split_mixed_shapely_array(data):
        field, geom_arr = construct_geometry_array(
            partial_geometry_array,
        )
        table = Table.from_arrays([geom_arr], schema=Schema([field]))
        layers.extend(_viz_geoarrow_table(table, **kwargs))

    return layers


def _viz_geo_interface(
    data: dict,
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    try:
        import pyarrow as pa
    except ImportError as e:
        raise ImportError(
            "pyarrow required for visualizing __geo_interface__ objects.\n"
            "Run `pip install pyarrow`.",
        ) from e

    try:
        import shapely
    except ImportError as e:
        raise ImportError(
            "shapely required for visualizing __geo_interface__ objects.\n"
            "Run `pip install shapely`.",
        ) from e

    if data["type"] in [
        "Point",
        "LineString",
        "Polygon",
        "MultiPoint",
        "MultiLineString",
        "MultiPolygon",
    ]:
        return _viz_shapely_scalar(shapely.from_geojson(json.dumps(data)), **kwargs)

    if data["type"] == "Feature":
        attribute_columns = {k: [v] for k, v in data["properties"].items()}
        table = pa.table(attribute_columns)
        shapely_geom = shapely.from_geojson(json.dumps(data["geometry"]))
        field, geom_arr = construct_geometry_array(np.array([shapely_geom]))
        return _viz_geoarrow_table(table.append_column(field, geom_arr), **kwargs)

    if data["type"] == "FeatureCollection":
        # We currently take a FeatureCollection through GeoPandas so that we can handle
        # mixed-geometry type collections
        import geopandas as gpd
        import pandas as pd

        check_pandas_version()

        attribute_columns_struct = pa.array(
            [feature["properties"] for feature in data["features"]],
        )

        fields = []
        arrays = []
        for field_idx in range(attribute_columns_struct.type.num_fields):
            fields.append(attribute_columns_struct.type.field(field_idx))
            arrays.append(struct_field(attribute_columns_struct, field_idx))  # type: ignore

        table = pa.Table.from_arrays(arrays, schema=pa.schema(fields))
        df = table.to_pandas(types_mapper=pd.ArrowDtype)

        shapely_geom_arr = shapely.from_geojson(
            [json.dumps(feature["geometry"]) for feature in data["features"]],
        )
        gdf = gpd.GeoDataFrame(df, geometry=shapely_geom_arr)  # type: ignore
        return _viz_geopandas_geodataframe(gdf, **kwargs)

    geo_interface_type = data["type"]
    raise ValueError(f"type '{geo_interface_type}' not supported.")


def _viz_geoarrow_chunked_array(
    ca: ChunkedArray,
    **kwargs: Any,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    field = ca.field.with_name("geometry")
    schema = Schema([field])
    table = Table.from_arrays([ca], schema=schema)
    table = add_positional_row_index(table)
    return _viz_geoarrow_table(table, **kwargs)


def _viz_geoarrow_table(
    table: Table,
    *,
    _viz_color: str,
    scatterplot_kwargs: ScatterplotLayerKwargs | None = None,
    path_kwargs: PathLayerKwargs | None = None,
    polygon_kwargs: PolygonLayerKwargs | None = None,
) -> list[ScatterplotLayer | PathLayer | PolygonLayer]:
    parsed_tables = parse_serialized_table(table)
    if len(parsed_tables) > 1:
        output: list[ScatterplotLayer | PathLayer | PolygonLayer] = []
        for parsed_table in parsed_tables:
            output.extend(
                _viz_geoarrow_table(
                    parsed_table,
                    _viz_color=_viz_color,
                    scatterplot_kwargs=scatterplot_kwargs,
                    path_kwargs=path_kwargs,
                    polygon_kwargs=polygon_kwargs,
                ),
            )

        return output
    table = parsed_tables[0]

    geometry_column_index = get_geometry_column_index(table.schema)
    assert geometry_column_index is not None, (
        "One column must have GeoArrow extension metadata"
    )

    geometry_field = table.schema.field(geometry_column_index)
    geometry_ext_type = geometry_field.metadata.get(b"ARROW:extension:name")

    if geometry_ext_type in [EXTENSION_NAME.POINT, EXTENSION_NAME.MULTIPOINT]:
        scatterplot_kwargs = scatterplot_kwargs if scatterplot_kwargs else {}

        if "get_fill_color" not in scatterplot_kwargs:
            scatterplot_kwargs["get_fill_color"] = _viz_color

        if "radius_min_pixels" not in scatterplot_kwargs:
            if len(table) <= 10_000:
                scatterplot_kwargs["radius_min_pixels"] = 2
            elif len(table) <= 100_000:
                scatterplot_kwargs["radius_min_pixels"] = 1
            elif len(table) <= 1_000_000:
                scatterplot_kwargs["radius_min_pixels"] = 0.5
            else:
                scatterplot_kwargs["radius_min_pixels"] = 0.2

        if "opacity" not in scatterplot_kwargs:
            if len(table) <= 10_000:
                scatterplot_kwargs["opacity"] = 0.9
            elif len(table) <= 100_000:
                scatterplot_kwargs["opacity"] = 0.7
            elif len(table) <= 1_000_000:
                scatterplot_kwargs["opacity"] = 0.5

        return [ScatterplotLayer(table=table, **scatterplot_kwargs)]

    if geometry_ext_type in [
        EXTENSION_NAME.LINESTRING,
        EXTENSION_NAME.MULTILINESTRING,
    ]:
        path_kwargs = path_kwargs if path_kwargs else {}

        if "get_color" not in path_kwargs:
            path_kwargs["get_color"] = _viz_color

        if "width_min_pixels" not in path_kwargs:
            if len(table) <= 1_000:
                path_kwargs["width_min_pixels"] = 1.5
            elif len(table) <= 10_000:
                path_kwargs["width_min_pixels"] = 1
            elif len(table) <= 100_000:
                path_kwargs["width_min_pixels"] = 0.7
            else:
                path_kwargs["width_min_pixels"] = 0.5

        if "opacity" not in path_kwargs:
            if len(table) <= 1_000:
                path_kwargs["opacity"] = 0.9
            elif len(table) <= 10_000:
                path_kwargs["opacity"] = 0.7
            elif len(table) <= 100_000:
                path_kwargs["opacity"] = 0.5

        return [PathLayer(table=table, **path_kwargs)]

    if geometry_ext_type in [
        EXTENSION_NAME.POLYGON,
        EXTENSION_NAME.MULTIPOLYGON,
        EXTENSION_NAME.BOX,
    ]:
        polygon_kwargs = polygon_kwargs if polygon_kwargs else {}

        if "get_fill_color" not in polygon_kwargs:
            polygon_kwargs["get_fill_color"] = _viz_color

        if "get_line_color" not in polygon_kwargs:
            polygon_kwargs["get_line_color"] = DEFAULT_POLYGON_LINE_COLOR

        if "opacity" not in polygon_kwargs:
            polygon_kwargs["opacity"] = 0.5

        if "line_width_min_pixels" not in polygon_kwargs:
            if len(table) <= 100:
                polygon_kwargs["line_width_min_pixels"] = 0.5
            if len(table) <= 1_000:
                polygon_kwargs["line_width_min_pixels"] = 0.45
            if len(table) <= 5_000:
                polygon_kwargs["line_width_min_pixels"] = 0.4
            elif len(table) <= 10_000:
                polygon_kwargs["line_width_min_pixels"] = 0.3
            elif len(table) <= 100_000:
                polygon_kwargs["line_width_min_pixels"] = 0.25
            else:
                polygon_kwargs["line_width_min_pixels"] = 0.2

        return [PolygonLayer(table=table, **polygon_kwargs)]

    raise ValueError(f"Unsupported extension type: '{geometry_ext_type!r}'.")



"""Helpers for viewport operations.

This is partially derived from pydeck at
(https://github.com/visgl/deck.gl/blob/63728ecbdaa2f99811900ec3709e5df0f9f8d228/bindings/pydeck/pydeck/data_utils/viewport_helpers.py)
under the Apache 2 license.
"""

# ruff: noqa: SLF001

from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any

from lonboard._geoarrow.ops.bbox import Bbox
from lonboard._geoarrow.ops.centroid import WeightedCentroid

if TYPE_CHECKING:
    from collections.abc import Sequence

    from lonboard._layer import BaseLayer


def get_bbox_center(layers: Sequence[BaseLayer]) -> tuple[Bbox, WeightedCentroid]:
    """Get the bounding box and geometric (weighted) center.

    Of all geometries in the table.
    """
    overall_bbox = Bbox()
    overall_centroid = WeightedCentroid()

    for layer in layers:
        overall_bbox.update(layer._bbox)
        overall_centroid.update(layer._weighted_centroid)

    return overall_bbox, overall_centroid


def bbox_to_zoom_level(bbox: Bbox) -> int:
    """Compute the zoom level of a bounding box.

    This is copied from pydeck: https://github.com/visgl/deck.gl/blob/63728ecbdaa2f99811900ec3709e5df0f9f8d228/bindings/pydeck/pydeck/data_utils/viewport_helpers.py#L125C1-L149C22

    Returns:
        Zoom level of map in a WGS84 Mercator projection

    """
    lat_diff = max(bbox.miny, bbox.maxy) - min(bbox.miny, bbox.maxy)
    lng_diff = max(bbox.minx, bbox.maxx) - min(bbox.minx, bbox.maxx)

    max_diff = max(lng_diff, lat_diff)
    zoom_level = None
    if max_diff < (360.0 / math.pow(2, 20)):
        zoom_level = 21
    else:
        zoom_level = int(
            -1
            * ((math.log(max_diff) / math.log(2.0)) - (math.log(360.0) / math.log(2))),
        )
        zoom_level = max(zoom_level, 1)

    return zoom_level


def compute_view(layers: Sequence[BaseLayer]) -> dict[str, Any]:
    """Automatically computes a view state for the data passed in."""
    bbox, center = get_bbox_center(layers)

    # When no geo column is found, bbox will have inf values
    try:
        zoom = bbox_to_zoom_level(bbox)
    except OverflowError:
        return {
            "longitude": center.x or 0,
            "latitude": center.y or 0,
            "zoom": 0,
            "pitch": 0,
            "bearing": 0,
        }
    else:
        return {
            "longitude": center.x,
            "latitude": center.y,
            "zoom": zoom,
            "pitch": 0,
            "bearing": 0,
        }



"""Import Arrow data from a C stream into a Table."""

from __future__ import annotations

from typing import TYPE_CHECKING

from arro3.core import ArrayReader, ChunkedArray, DataType, Table

if TYPE_CHECKING:
    from arro3.core.types import ArrowStreamExportable


def import_arrow_c_stream(stream: ArrowStreamExportable) -> Table | ChunkedArray:
    """Import an Arrow stream into a Table.

    If the stream is a struct but **not** a geoarrow extension type, then we assume it's
    table input. Otherwise, we assume it's an array input, which we then coerce into a
    Table with a single field named "geometry".
    """
    array_reader = ArrayReader.from_arrow(stream)

    if DataType.is_struct(
        array_reader.field.type,
    ) and not array_reader.field.metadata_str.get(
        "ARROW:extension:name",
        "",
    ).startswith("geoarrow"):
        # If the field is a struct but **not** a geoarrow extension type, then we assume
        # it's table input.
        return Table.from_arrow(array_reader)

    return ChunkedArray.from_arrow(array_reader)



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any, TypeVar

import arro3.compute as ac
import numpy as np
from arro3.core import ChunkedArray, DataType, Scalar, Schema, list_flatten

from lonboard._compat import check_pandas_version
from lonboard._constants import EXTENSION_NAME, MIN_INTEGER_FLOAT32

if TYPE_CHECKING:
    from collections.abc import Sequence

    import geopandas as gpd
    import pandas as pd
    from numpy.typing import NDArray

    from lonboard._base import BaseExtension

    DF = TypeVar("DF", bound=pd.DataFrame)

GEOARROW_EXTENSION_TYPE_NAMES = {e.value for e in EXTENSION_NAME}


def get_geometry_column_index(schema: Schema) -> int | None:
    """Get the positional index of the geometry column in a pyarrow Schema."""
    field_idxs = []

    for field_idx in range(len(schema)):
        field_metadata = schema.field(field_idx).metadata
        if (
            field_metadata
            and field_metadata.get(b"ARROW:extension:name")
            in GEOARROW_EXTENSION_TYPE_NAMES
        ):
            field_idxs.append(field_idx)

    if len(field_idxs) > 1:
        raise ValueError("Multiple geometry columns not supported.")
    if len(field_idxs) == 1:
        return field_idxs[0]
    return None


def auto_downcast(df: DF) -> DF:
    """Automatically downcast types to smallest data size.

    Args:
        df: pandas DataFrame or geopandas GeoDataFrame

    Returns:
        DataFrame with downcasted data types

    """
    import pandas as pd

    check_pandas_version()

    # Convert objects to numeric types where possible.
    # Note: we have to exclude geometry because
    # `convert_dtypes(dtype_backend="pyarrow")` fails on the geometory column, but we
    # also have to manually cast to a non-geo data frame because it'll fail to convert
    # dtypes on a GeoDataFrame without a geom col
    casted_df = pd.DataFrame(df.select_dtypes(exclude="geometry")).convert_dtypes(  # type: ignore
        infer_objects=True,
        convert_string=True,
        convert_integer=True,
        convert_boolean=True,
        convert_floating=True,
        dtype_backend="pyarrow",
    )
    df[casted_df.columns] = casted_df

    # Try to convert _all_ integer columns to unsigned integer columns, but use
    # errors='ignore' to return signed integer data types for columns with negative
    # integers.
    for col_name in df.select_dtypes(np.integer).columns:  # type: ignore
        with contextlib.suppress(Exception):
            df[col_name] = pd.to_numeric(
                df[col_name],
                downcast="unsigned",
                dtype_backend="pyarrow",
            )

    # For any integer columns that are still signed integer, downcast those to smaller
    # signed types
    for col_name in df.select_dtypes(np.signedinteger).columns:  # type: ignore
        with contextlib.suppress(Exception):
            df[col_name] = pd.to_numeric(
                df[col_name],
                downcast="signed",
                dtype_backend="pyarrow",
            )

    for col_name in df.select_dtypes(np.floating).columns:  # type: ignore
        with contextlib.suppress(Exception):
            df[col_name] = pd.to_numeric(
                df[col_name],
                downcast="float",
                dtype_backend="pyarrow",
            )

    return df


def remove_extension_kwargs(
    extensions: Sequence[BaseExtension],
    kwargs: dict[str, Any],
) -> dict[str, Any]:
    """Remove extension properties from kwargs, returning the removed properties.

    **This mutates the kwargs input.**
    """
    extension_kwargs: dict[str, Any] = {}
    if extensions:
        for extension in extensions:
            for extension_prop_name in extension._layer_traits:  # noqa: SLF001
                if extension_prop_name in kwargs:
                    extension_kwargs[extension_prop_name] = kwargs.pop(
                        extension_prop_name,
                    )

    return extension_kwargs


def split_mixed_gdf(gdf: gpd.GeoDataFrame) -> list[gpd.GeoDataFrame]:
    """Split a GeoDataFrame into one or more GeoDataFrames with unique geometry type."""
    indices = indices_by_geometry_type(gdf.geometry)
    if indices is None:
        return [gdf]

    point_indices, linestring_indices, polygon_indices = indices

    # Here we intentionally check geometries in a specific order.
    # Starting from polygons, then linestrings, then points,
    # so that the order of generated layers is polygon, then path then scatterplot.
    # This ensures that points are rendered on top and polygons on the bottom.
    gdfs = []
    for single_type_geometry_indices in (
        polygon_indices,
        linestring_indices,
        point_indices,
    ):
        if len(single_type_geometry_indices) > 0:
            gdfs.append(gdf.iloc[single_type_geometry_indices])

    return gdfs


def split_mixed_shapely_array(
    geometry: NDArray[np.object_],
) -> list[NDArray[np.object_]]:
    """Split a shapely array into one or more arrays with unique geometry type."""
    indices = indices_by_geometry_type(geometry)
    if indices is None:
        return [geometry]

    point_indices, linestring_indices, polygon_indices = indices

    # Here we intentionally check geometries in a specific order.
    # Starting from polygons, then linestrings, then points,
    # so that the order of generated layers is polygon, then path then scatterplot.
    # This ensures that points are rendered on top and polygons on the bottom.
    arrays = []
    for single_type_geometry_indices in (
        polygon_indices,
        linestring_indices,
        point_indices,
    ):
        if len(single_type_geometry_indices) > 0:
            arrays.append(geometry[single_type_geometry_indices])

    return arrays


def indices_by_geometry_type(
    geometry: NDArray[np.object_],
) -> tuple[NDArray[np.int64], NDArray[np.int64], NDArray[np.int64]] | None:
    import shapely
    from shapely import GeometryType

    type_ids = np.array(shapely.get_type_id(geometry))
    unique_type_ids = set(np.unique(type_ids))

    if GeometryType.GEOMETRYCOLLECTION in unique_type_ids:
        raise ValueError("GeometryCollections not currently supported")

    if GeometryType.LINEARRING in unique_type_ids:
        raise ValueError("LinearRings not currently supported")

    if len(unique_type_ids) == 1:
        return None

    if len(unique_type_ids) == 2:
        if unique_type_ids == {GeometryType.POINT, GeometryType.MULTIPOINT}:
            return None

        if unique_type_ids == {GeometryType.LINESTRING, GeometryType.MULTILINESTRING}:
            return None

        if unique_type_ids == {GeometryType.POLYGON, GeometryType.MULTIPOLYGON}:
            return None

    point_indices = np.where(
        (type_ids == GeometryType.POINT) | (type_ids == GeometryType.MULTIPOINT),
    )[0]

    linestring_indices = np.where(
        (type_ids == GeometryType.LINESTRING)
        | (type_ids == GeometryType.MULTILINESTRING),
    )[0]

    polygon_indices = np.where(
        (type_ids == GeometryType.POLYGON) | (type_ids == GeometryType.MULTIPOLYGON),
    )[0]

    return point_indices, linestring_indices, polygon_indices


def timestamp_start_offset(timestamps: ChunkedArray) -> int:
    timestamps = timestamps.cast(DataType.list(DataType.int64()))

    min_timestamp = ac.min(list_flatten(timestamps))
    return MIN_INTEGER_FLOAT32 - min_timestamp.as_py()


def timestamp_max_physical_value(timestamps: ChunkedArray) -> int:
    # Cast to int64 type
    timestamps = timestamps.cast(DataType.list(DataType.int64()))

    min_timestamp = ac.min(list_flatten(timestamps))
    max_timestamp = ac.max(list_flatten(timestamps))
    start_offset_adjustment = Scalar(
        MIN_INTEGER_FLOAT32 - min_timestamp.as_py(),
        type=DataType.int64(),
    )
    return start_offset_adjustment.as_py() + max_timestamp.as_py()



"""Experimental layers for lonboard.

These layers have not been as well tested as other layers. You may encounter crashes or
unexpected behavior when using them.
"""

from ._layer import ArcLayer, TextLayer, TripsLayer



# ruff: noqa

from __future__ import annotations

import math
import warnings
from typing import TYPE_CHECKING, Any

import arro3.compute as ac
from arro3.core import (
    Array,
    ChunkedArray,
    DataType,
    Scalar,
    Table,
    list_array,
    list_flatten,
    list_offsets,
)

from lonboard._constants import MAX_INTEGER_FLOAT32, MIN_INTEGER_FLOAT32
from lonboard._serialization import TIMESTAMP_ACCESSOR_SERIALIZATION
from lonboard._utils import get_geometry_column_index
from lonboard.traits import FixedErrorTraitType

if TYPE_CHECKING:
    from traitlets.traitlets import TraitType

    from lonboard._layer import BaseArrowLayer


class TimestampAccessor(FixedErrorTraitType):
    """A representation of a deck.gl coordinate-timestamp accessor.

    deck.gl handles timestamps on the GPU as float32 values. This class will validate
    that the input timestamps are representable as float32 integers, and will
    automatically reduce the precision of input data if necessary to fit inside a
    float32.

    Accepted input includes:

    - A pyarrow [`ListArray`][pyarrow.ListArray] containing a temporal array such as a
        TimestampArray. Each value in the array will be used as the value for the object
        at the same row index.
    - Any Arrow list array containing a temporal array from a library that implements
      the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = None
    info_text = "a Arrow ListArray representing a nested array of timestamps"

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **TIMESTAMP_ACCESSOR_SERIALIZATION)

    def _reduce_precision(
        self,
        obj: BaseArrowLayer,
        value: ChunkedArray,
    ) -> ChunkedArray:
        # First, find the "spread" of existing values: the range between min and max of
        # the existing timestamps.
        min_val = ac.min(list_flatten(value))
        max_val = ac.max(list_flatten(value))

        # deck.gl stores timestamps as float32. Therefore, we need to validate that the
        # actual range of our data fits into the float32 integer range.
        actual_spread = ac.sub(max_val, min_val).cast(DataType.int64())[0].as_py()
        acceptable_spread = MAX_INTEGER_FLOAT32 - MIN_INTEGER_FLOAT32

        # If if already fits into the float32 integer range, we're done.
        if actual_spread <= acceptable_spread:
            return value

        # Otherwise, we need to reduce the precision of our data. We can do this by
        # changing the time unit of our data. So if our timestamps start with nanosecond
        # precision, we can downcast our data to microsecond, millisecond, or second
        # precision.
        # First, we want to figure out how many digits of precision to remove.
        digits_to_remove = math.log10(actual_spread / acceptable_spread)

        # Access the existing time unit and timezone
        value_type = value.type.value_type
        assert value_type is not None
        time_unit = value_type.time_unit
        tz = value_type.tz

        # We can only reduce precision in orders of 1,000.
        if digits_to_remove <= 3:
            digits_to_remove = 3
        elif digits_to_remove <= 6:
            digits_to_remove = 6
        elif digits_to_remove <= 9:
            digits_to_remove = 9
        else:
            # Cause the exception to be raised in the next block
            digits_to_remove = 999999

        # Validate that the precision reduction is possible. E.g. we can't reduce the
        # precision of second timestamps because we have no way to represent timestamps
        # larger than a second.
        if (
            (time_unit == "ns" and digits_to_remove > 9)
            or (time_unit == "us" and digits_to_remove > 6)
            or (time_unit == "ms" and digits_to_remove > 3)
            or (time_unit == "s")
        ):
            self.error(
                obj,
                value,
                info=(
                    "The range of the timestamp column cannot is larger than what "
                    "can be represented at second precision in a float32. deck.gl "
                    "uses float32 for the timestamps in the TripsLayer. "
                    "Choose a smaller temporal subset of data."
                ),
            )

        # Figure out the new data type given the existing data type and the number of
        # digits of precision we're removing.
        if time_unit == "ns":
            if digits_to_remove == 3:
                new_data_type = DataType.timestamp("us", tz=tz)
            elif digits_to_remove == 6:
                new_data_type = DataType.timestamp("ms", tz=tz)
            elif digits_to_remove == 9:
                new_data_type = DataType.timestamp("s", tz=tz)
            else:
                assert False
        elif time_unit == "us":
            if digits_to_remove == 3:
                new_data_type = DataType.timestamp("ms", tz=tz)
            elif digits_to_remove == 6:
                new_data_type = DataType.timestamp("s", tz=tz)
            else:
                assert False
        elif time_unit == "ms":
            if digits_to_remove == 3:
                new_data_type = DataType.timestamp("s", tz=tz)
            else:
                assert False
        else:
            assert False

        new_time_unit = new_data_type.time_unit
        warnings.warn(
            f"Reducing precision of input timestamp data to '{new_time_unit}'"
            " to fit into available GPU precision.",
        )

        # Actually reduce the precision of each chunk of the input data, assigning the
        # new data type
        offsets_reader = list_offsets(value)
        inner_values_reader = list_flatten(value)

        divisor = Scalar(int(math.pow(10, digits_to_remove)), type=DataType.int64())

        reduced_precision_chunks = []
        for offsets, inner_values in zip(offsets_reader, inner_values_reader):
            reduced_precision_values = ac.div(
                inner_values.cast(DataType.int64()),
                divisor,
            )
            reduced_precision_chunks.append(
                list_array(offsets, reduced_precision_values.cast(new_data_type)),
            )

        return ChunkedArray(reduced_precision_chunks)

    def _validate_timestamp_offsets(self, obj: BaseArrowLayer, value: ChunkedArray):
        """Validate that the offsets of the list array used for the timestamp column match
        the offsets of the list array used for the LineString array in the geometry
        column.
        """
        table: Table = obj.table
        geom_idx = get_geometry_column_index(table.schema)
        assert geom_idx is not None
        geom_col = table.column(geom_idx)

        # TODO: this chunking may not match depending on whether the table has already
        # been rechunked by this point.
        for geom_chunk, timestamp_chunk in zip(geom_col.chunks, value.chunks):
            geom_offsets = list_offsets(geom_chunk)
            timestamp_offsets = list_offsets(timestamp_chunk)
            if geom_offsets != timestamp_offsets:
                self.error(
                    obj,
                    value,
                    info="timestamp array's offsets to match geometry array's offsets.",
                )

    def validate(self, obj: BaseArrowLayer, value) -> ChunkedArray:
        if hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_list(value.type):
            self.error(obj, value, info="timestamp array to be a list-type array")

        value_type = value.type.value_type
        assert value_type is not None
        if not DataType.is_temporal(value_type):
            self.error(obj, value, info="timestamp array to have a temporal child.")

        value = self._reduce_precision(obj, value)
        value = value.rechunk(max_chunksize=obj._rows_per_chunk)
        self._validate_timestamp_offsets(obj, value)
        return value



from typing import Any, ClassVar

import traitlets
from anywidget import AnyWidget
from ipywidgets import Widget

msg = """
Unexpected keyword argument: '{provided_trait_name}'.
Check the spelling of your parameters. If you're trying to use layer properties added by
a layer extension, ensure you've passed the extension object into the `extensions`
parameter of the layer.
"""


class BaseWidget(Widget):
    def __init__(self, **kwargs: Any) -> None:
        # Raise error on unknown keyword name
        # Note: we don't use `class_own_traits()` because some layer props are set on
        # BaseLayer
        layer_trait_names = self.trait_names()
        for provided_trait_name in kwargs:
            if provided_trait_name not in layer_trait_names:
                raise TypeError(msg.format(provided_trait_name=provided_trait_name))

        super().__init__(**kwargs)


class BaseAnyWidget(AnyWidget):
    def __init__(self, **kwargs: Any) -> None:
        # Raise error on unknown keyword name
        layer_trait_names = self.trait_names()
        for provided_trait_name in kwargs:
            if provided_trait_name not in layer_trait_names:
                raise TypeError(msg.format(provided_trait_name=provided_trait_name))

        super().__init__(**kwargs)


class BaseExtension(BaseWidget):
    _extension_type: traitlets.Unicode

    _layer_traits: ClassVar[dict[str, traitlets.TraitType]] = {}
    """Traits from this extension to dynamically assign onto a layer."""



from __future__ import annotations

import json
import webbrowser
from pathlib import Path
from tempfile import NamedTemporaryFile

import click
from arro3.core import Table
from pyproj import CRS

from lonboard import viz
from lonboard._constants import EXTENSION_NAME


def read_pyogrio(path: Path) -> Table:
    """Read path using pyogrio and convert field metadata to geoarrow.

    Args:
        path: Path to file readable by pyogrio

    """
    try:
        from pyogrio.raw import open_arrow
    except ImportError as e:
        raise ImportError(
            "pyogrio is a required dependency for the CLI for reading data sources \n"
            "other than GeoParquet.\n"
            "Install with `pip install pyogrio`.",
        ) from e

    with open_arrow(path, use_pyarrow=False) as source:
        meta, stream = source
        table = Table.from_arrow(stream)

    # The `geometry_name` key always exists but can be an empty string. In the case of
    # an empty string, we want to default to `wkb_geometry`
    geometry_column_name = meta.get("geometry_name") or "wkb_geometry"
    # TODO: assert there are not two column names of wkb_geometry, nor an existing
    # column named "geometry"

    # Rename wkb_geometry to geometry
    geometry_column_index = [
        i for (i, name) in enumerate(table.column_names) if name == geometry_column_name
    ][0]

    schema = table.schema
    field = schema.field(geometry_column_index)

    metadata: dict[bytes, bytes] = field.metadata
    if metadata.get(b"ARROW:extension:name") == EXTENSION_NAME.OGC_WKB:
        # Parse CRS and create PROJJSON
        ext_meta = {"crs": CRS.from_user_input(meta["crs"]).to_json_dict()}

        # Replace metadata
        metadata[b"ARROW:extension:name"] = EXTENSION_NAME.WKB
        metadata[b"ARROW:extension:metadata"] = json.dumps(ext_meta).encode()

    new_field = field.with_name("geometry").with_metadata(metadata)
    new_schema = schema.set(geometry_column_index, new_field)
    return table.with_schema(new_schema)


def read_parquet(path: Path) -> tuple[Table, dict]:
    """Read Parquet file using either pyarrow or arro3.

    arro3.io.read_parquet is not multi-threaded (as of arro3 0.2.1), so pyarrow can be
    up to 4x faster on an 8-core machine. Because of this, we prefer pyarrow if it's
    installed, and fall back to arro3 otherwise.

    Args:
        path: path to Parquet file.

    Raises:
        ValueError: if there's no GeoParquet metadata in the file

    Returns:
        arro3 Table

    """
    try:
        import pyarrow.parquet as pq

    except ImportError:
        from arro3.io import read_parquet

        reader = read_parquet(path)

        if "geo" not in reader.schema.metadata_str:
            raise ValueError("Expected geo metadata in Parquet file")  # noqa: B904

        table = reader.read_all()
        geo_meta = json.loads(table.schema.metadata_str["geo"])

        return table, geo_meta
    else:
        file = pq.ParquetFile(path)
        if b"geo" not in file.metadata.metadata:
            raise ValueError("Expected geo metadata in Parquet file")
        geo_meta = json.loads(file.metadata.metadata.get(b"geo"))

        table = Table.from_arrow(file.read())

        return table, geo_meta


def read_geoparquet(path: Path) -> Table:
    """Read GeoParquet file at path using pyarrow or arro3.io.

    Args:
        path: Path to GeoParquet file

    """
    table, geo_meta = read_parquet(path)
    geometry_column_name = geo_meta["primary_column"]
    geometry_column_index = [
        i for (i, name) in enumerate(table.schema.names) if name == geometry_column_name
    ][0]

    metadata: dict[bytes, bytes] = {
        b"ARROW:extension:name": EXTENSION_NAME.WKB,
    }
    crs_dict = geo_meta["columns"][geometry_column_name].get("crs")
    if crs_dict:
        # Parse CRS and create PROJJSON
        ext_meta = {"crs": crs_dict}
        metadata[b"ARROW:extension:metadata"] = json.dumps(ext_meta).encode()

    new_field = table.schema.field(geometry_column_index).with_metadata(metadata)
    new_schema = table.schema.set(geometry_column_index, new_field)
    return table.with_schema(new_schema)


@click.command()
@click.option(
    "-o",
    "--output",
    type=click.Path(path_type=Path),
    default=None,
    help=(
        "The output path for the generated HTML file. "
        "If not provided, will save to a temporary file."
    ),
)
@click.option(
    "--open/--no-open",
    "open_browser",
    default=None,
    help=(
        "Whether to open a web browser tab with the generated map. "
        "By default, the web browser is not opened when --output is provided, "
        "but is in other cases."
    ),
)
@click.argument("files", nargs=-1, type=click.Path(path_type=Path))
def main(
    output: Path | None,
    open_browser: bool | None,  # noqa: FBT001
    files: list[Path],
) -> None:
    """Interactively visualize geospatial data using Lonboard.

    This CLI can be used either to quickly view local files or to create static HTML
    files.
    """
    tables = []
    for path in files:
        if path.suffix == ".parquet":
            table = read_geoparquet(path)
        else:
            table = read_pyogrio(path)

        tables.append(table)

    map_ = viz(tables)

    # If -o flag passed, write to file; otherwise write to temporary file
    if output:
        map_.to_html(output)

        # Default to not opening browser when None
        if open_browser is True:
            webbrowser.open_new_tab(f"file://{output.absolute()}")
    else:
        with NamedTemporaryFile("wt", suffix=".html", delete=False) as f:
            map_.to_html(f)

            # Default to opening browser when None
            if open_browser is not False:
                webbrowser.open_new_tab(f"file://{f.name}")



from __future__ import annotations

import json
import re
from typing import TYPE_CHECKING

import numpy as np
from arro3.core import (
    Array,
    ChunkedArray,
    Field,
    Table,
    fixed_size_list_array,
    list_array,
    struct_field,
)

from lonboard._constants import EXTENSION_NAME

if TYPE_CHECKING:
    import duckdb
    import pyproj

DUCKDB_SPATIAL_TYPES = {
    "GEOMETRY",
    "WKB_BLOB",
    "POINT_2D",
    "LINESTRING_2D",
    "POLYGON_2D",
    "BOX_2D",
}


def from_duckdb(
    rel: duckdb.DuckDBPyRelation,
    *,
    crs: str | pyproj.CRS | None = None,
) -> Table:
    geom_col_idxs = [
        i for i, t in enumerate(rel.types) if str(t) in DUCKDB_SPATIAL_TYPES
    ]

    if len(geom_col_idxs) == 0:
        raise ValueError("No geometry column found in query.")

    if len(geom_col_idxs) > 1:
        msg = "Multiple geometry columns found in query.\n\n"
        for geom_col_idx in geom_col_idxs:
            msg += (
                f"- name: {rel.columns[geom_col_idx]}, "
                f"type: {rel.types[geom_col_idx]} \n"
            )

        raise ValueError(msg)

    geom_col_idx = geom_col_idxs[0]
    geom_type = rel.types[geom_col_idx]
    if geom_type == "WKB_BLOB":
        return _from_geoarrow(
            rel,
            extension_type=EXTENSION_NAME.WKB,
            geom_col_idx=geom_col_idx,
            crs=crs,
        )
    if geom_type == "GEOMETRY":
        return _from_geometry(rel, geom_col_idx=geom_col_idx, crs=crs)
    if geom_type == "POINT_2D":
        return _from_geoarrow(
            rel,
            extension_type=EXTENSION_NAME.POINT,
            geom_col_idx=geom_col_idx,
            crs=crs,
        )
    if geom_type == "LINESTRING_2D":
        return _from_geoarrow(
            rel,
            extension_type=EXTENSION_NAME.LINESTRING,
            geom_col_idx=geom_col_idx,
            crs=crs,
        )
    if geom_type == "POLYGON_2D":
        return _from_geoarrow(
            rel,
            extension_type=EXTENSION_NAME.POLYGON,
            geom_col_idx=geom_col_idx,
            crs=crs,
        )
    if geom_type == "BOX_2D":
        return _from_box2d(
            rel,
            geom_col_idx=geom_col_idx,
            crs=crs,
        )
    raise ValueError(f"Unsupported geometry type: {geom_type}")


def _from_geometry(
    rel: duckdb.DuckDBPyRelation,
    *,
    geom_col_idx: int,
    crs: str | pyproj.CRS | None = None,
) -> Table:
    from duckdb import ColumnExpression, FunctionExpression

    other_col_names = [name for i, name in enumerate(rel.columns) if i != geom_col_idx]
    if other_col_names:
        non_geo_table = Table.from_arrow(rel.select(*other_col_names).arrow())
    else:
        non_geo_table = None
    geom_col_name = rel.columns[geom_col_idx]

    # A poor-man's string interpolation check
    # We can't pass in SQL-templated strings for the column name
    re_match = r"[a-zA-Z][a-zA-Z0-9_]*"
    assert re.fullmatch(
        re_match,
        geom_col_name,
    ), f"Expected geometry column name to match regex: {re_match}"

    geom_table = Table.from_arrow(
        rel.select(
            FunctionExpression("st_aswkb", ColumnExpression(geom_col_name)).alias(
                geom_col_name,
            ),
        ).arrow(),
    )

    metadata = _make_geoarrow_field_metadata(EXTENSION_NAME.WKB, crs)
    geom_field = geom_table.schema.field(0).with_metadata(metadata)
    if non_geo_table is not None:
        return non_geo_table.append_column(geom_field, geom_table.column(0))
    # Need to set geospatial metadata onto the Arrow table, because the table
    # returned from duckdb has none.
    new_schema = geom_table.schema.set(0, geom_field)
    return geom_table.with_schema(new_schema)


def _from_geoarrow(
    rel: duckdb.DuckDBPyRelation,
    *,
    extension_type: EXTENSION_NAME,
    geom_col_idx: int,
    crs: str | pyproj.CRS | None = None,
) -> Table:
    table = Table.from_arrow(rel.arrow())
    metadata = _make_geoarrow_field_metadata(extension_type, crs)
    geom_field = table.schema.field(geom_col_idx).with_metadata(metadata)
    return table.set_column(geom_col_idx, geom_field, table.column(geom_col_idx))


def _from_box2d(
    rel: duckdb.DuckDBPyRelation,
    *,
    geom_col_idx: int,
    crs: str | pyproj.CRS | None = None,
) -> Table:
    table = Table.from_arrow(rel.arrow())
    geom_col = table.column(geom_col_idx)

    polygon_chunks: list[Array] = []
    for geom_chunk in geom_col.chunks:
        polygon_array = _convert_box2d_to_geoarrow_polygon_array(geom_chunk)
        polygon_chunks.append(polygon_array)

    metadata = _make_geoarrow_field_metadata(EXTENSION_NAME.POLYGON, crs)
    prev_field = table.schema.field(geom_col_idx)
    geom_field = Field(prev_field.name, polygon_chunks[0].type, metadata=metadata)
    return table.set_column(geom_col_idx, geom_field, ChunkedArray(polygon_chunks))


def _convert_box2d_to_geoarrow_polygon_array(
    geom_col: Array,
) -> Array:
    """Manual conversion of the duckdb box_2d type to a GeoArrow Polygon array.

    We don't wish to add dependencies so we reimplement this here using numpy.
    """
    # Extract the bounding box columns from the Arrow struct
    # NOTE: this assumes that the box ordering is minx, miny, maxx, maxy
    # Note sure whether the positional ordering or the named fields is more stable
    min_x = struct_field(geom_col, 0)
    min_y = struct_field(geom_col, 1)
    max_x = struct_field(geom_col, 2)
    max_y = struct_field(geom_col, 3)

    # Provision memory for the output coordinates. For closed polygons, each input box
    # becomes 5 coordinates.
    # The first dimension is length and the second dimension allows for x and y
    coords = np.zeros(((len(geom_col) * 5), 2), dtype=np.float64)

    # Compute the insertion indexes for each corner of the box.
    # Start at bottom left
    bottom_left_idxs = np.arange(0, len(geom_col) * 5, 5)
    bottom_right_idxs = bottom_left_idxs + 1
    upper_right_idxs = bottom_right_idxs + 1
    upper_left_idxs = upper_right_idxs + 1
    bottom_left_idxs_again = upper_left_idxs + 1

    # Insert each corner into the coordinate array at these specified indexes
    coords[bottom_left_idxs, :] = np.column_stack([min_x, min_y])
    coords[bottom_right_idxs, :] = np.column_stack([max_x, min_y])
    coords[upper_right_idxs, :] = np.column_stack([max_x, max_y])
    coords[upper_left_idxs, :] = np.column_stack([min_x, max_y])
    coords[bottom_left_idxs_again, :] = np.column_stack([min_x, min_y])

    # Create the geoarrow Polygon offset arrays
    # Each ring has 5 coordinates
    ring_offsets = np.arange(0, (len(geom_col) + 1) * 5, 5, dtype=np.int32)
    # Each geometry has a single ring
    geom_offsets = np.arange(0, len(ring_offsets), dtype=np.int32)

    # Construct the final PolygonArray
    coords = fixed_size_list_array(coords.ravel("C"), 2)
    ring_array = list_array(ring_offsets, coords)
    polygon_array = list_array(geom_offsets, ring_array)
    return polygon_array  # noqa: RET504


# TODO: refactor, put helper in lonboard._geoarrow.crs?
def _make_geoarrow_field_metadata(
    extension_type: EXTENSION_NAME,
    crs: str | pyproj.CRS | None = None,
) -> dict[bytes, bytes]:
    import pyproj

    metadata: dict[bytes, bytes] = {b"ARROW:extension:name": extension_type}

    if crs is not None:
        if isinstance(crs, str):
            crs = pyproj.CRS.from_user_input(crs)

        ext_meta = {"crs": crs.to_json_dict()}
        metadata[b"ARROW:extension:metadata"] = json.dumps(ext_meta).encode()

    return metadata



from __future__ import annotations

import json
from enum import Enum
from typing import TYPE_CHECKING

import numpy as np
from arro3.core import Array, DataType, Field, fixed_size_list_array, list_array

from lonboard._geoarrow.crs import serialize_crs

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import NDArray
    from pyproj import CRS


class CoordinateDimension(str, Enum):
    XY = "xy"
    XYZ = "xyz"
    XYM = "xym"
    XYZM = "xyzm"


def coord_storage_type(*, interleaved: bool, dims: CoordinateDimension) -> DataType:
    """Generate the storage type of a geoarrow coordinate array.

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions

    """
    if interleaved:
        return DataType.list(Field(dims, DataType.float64()), len(dims))

    if dims == CoordinateDimension.XY:
        return DataType.struct(
            [
                Field("x", DataType.float64()),
                Field("y", DataType.float64()),
            ],
        )
    if dims == CoordinateDimension.XYZ:
        return DataType.struct(
            [
                Field("x", DataType.float64()),
                Field("y", DataType.float64()),
                Field("z", DataType.float64()),
            ],
        )
    if dims == CoordinateDimension.XYM:
        return DataType.struct(
            [
                Field("x", DataType.float64()),
                Field("y", DataType.float64()),
                Field("m", DataType.float64()),
            ],
        )
    if dims == CoordinateDimension.XYZM:
        return DataType.struct(
            [
                Field("x", DataType.float64()),
                Field("y", DataType.float64()),
                Field("z", DataType.float64()),
                Field("m", DataType.float64()),
            ],
        )

    raise ValueError("Unreachable")


def linestring_storage_type(
    *,
    interleaved: bool,
    dims: CoordinateDimension,
    large_list: bool = False,
) -> DataType:
    """Generate the storage type of a geoarrow.linestring array.

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type

    """
    vertices_type = coord_storage_type(interleaved=interleaved, dims=dims)
    if large_list:
        return DataType.large_list(Field("vertices", vertices_type))
    return DataType.list(Field("vertices", vertices_type))


def polygon_storage_type(
    *,
    interleaved: bool,
    dims: CoordinateDimension,
    large_list: bool = False,
) -> DataType:
    """Generate the storage type of a geoarrow.polygon array.

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type

    """
    rings_type = linestring_storage_type(
        large_list=large_list,
        interleaved=interleaved,
        dims=dims,
    )
    if large_list:
        return DataType.large_list(Field("rings", rings_type))
    return DataType.list(Field("rings", rings_type))


def multipoint_storage_type(
    *,
    interleaved: bool,
    dims: CoordinateDimension,
    large_list: bool = False,
) -> DataType:
    """Generate the storage type of a geoarrow.multipoint array.

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type

    """
    points_type = coord_storage_type(interleaved=interleaved, dims=dims)
    if large_list:
        return DataType.large_list(Field("points", points_type))
    return DataType.list(Field("points", points_type))


def multilinestring_storage_type(
    *,
    interleaved: bool,
    dims: CoordinateDimension,
    large_list: bool = False,
) -> DataType:
    """Generate the storage type of a geoarrow.multilinestring array.

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type

    """
    linestrings_type = linestring_storage_type(
        large_list=large_list,
        interleaved=interleaved,
        dims=dims,
    )
    if large_list:
        return DataType.large_list(Field("linestrings", linestrings_type))
    return DataType.list(Field("linestrings", linestrings_type))


def multipolygon_storage_type(
    *,
    interleaved: bool,
    dims: CoordinateDimension,
    large_list: bool = False,
) -> DataType:
    """Generate the storage type of a geoarrow.multipolygon array.

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type

    """
    polygons_type = polygon_storage_type(
        large_list=large_list,
        interleaved=interleaved,
        dims=dims,
    )
    if large_list:
        return DataType.large_list(Field("polygons", polygons_type))
    return DataType.list(Field("polygons", polygons_type))


def offsets_to_arrow(
    offsets: tuple[NDArray[np.int32], ...] | tuple[NDArray[np.int64], ...],
) -> tuple[Sequence[Array], bool]:
    """Return a tuple of Arrow arrays from offsets.

    Returns:
        Arrays; whether they're large.

    """
    # Shapely historically produced int64 offset arrays in `to_ragged_array`, but as of
    # a recent version (2.1? 2.1.1?) switched to producing `int32` arrays where
    # possible. In the case that we receive `int64` arrays, we downcast them to int32 if
    # possible
    if all(offset_arr.dtype == np.int32 for offset_arr in offsets):
        return [Array(offset_arr) for offset_arr in offsets], False

    if any(offset_arr[-1] >= np.iinfo(np.int32).max for offset_arr in offsets):
        return [Array(offset_arr) for offset_arr in offsets], True

    return [Array(offset_arr.astype(np.int32)) for offset_arr in offsets], False


def construct_geometry_array(  # noqa: PLR0915
    shapely_arr: NDArray[np.object_],
    include_z: bool | None = None,  # noqa: FBT001
    *,
    field_name: str = "geometry",
    crs: CRS | None = None,
) -> tuple[Field, Array]:
    import shapely
    from shapely import GeometryType

    geom_type, coords, np_offsets = shapely.to_ragged_array(
        shapely_arr,
        include_z=include_z,
    )
    offsets, is_large_offset = offsets_to_arrow(np_offsets)

    if coords.shape[-1] == 2:
        dims = CoordinateDimension.XY
    elif coords.shape[-1] == 3:
        dims = CoordinateDimension.XYZ
    else:
        raise ValueError(f"Unexpected coords dimensions: {coords.shape}")

    extension_metadata: dict[str, str] = {}
    if crs is not None:
        extension_metadata["ARROW:extension:metadata"] = json.dumps(serialize_crs(crs))

    if geom_type == GeometryType.POINT:
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims)).cast(
            coord_storage_type(interleaved=True, dims=dims),
        )
        extension_metadata["ARROW:extension:name"] = "geoarrow.point"
        field = Field(
            field_name,
            arrow_coords.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_coords

    if geom_type == GeometryType.LINESTRING:
        assert len(offsets) == 1, "Expected one offsets array"
        (geom_offsets,) = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_geoms = list_array(geom_offsets, arrow_coords).cast(
            linestring_storage_type(
                interleaved=True,
                dims=dims,
                large_list=is_large_offset,
            ),
        )
        extension_metadata["ARROW:extension:name"] = "geoarrow.linestring"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    if geom_type == GeometryType.POLYGON:
        assert len(offsets) == 2, "Expected two offsets arrays"
        ring_offsets, geom_offsets = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_rings = list_array(ring_offsets, arrow_coords)
        arrow_geoms = list_array(geom_offsets, arrow_rings).cast(
            polygon_storage_type(
                interleaved=True,
                dims=dims,
                large_list=is_large_offset,
            ),
        )
        extension_metadata["ARROW:extension:name"] = "geoarrow.polygon"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    if geom_type == GeometryType.MULTIPOINT:
        assert len(offsets) == 1, "Expected one offsets array"
        (geom_offsets,) = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_geoms = list_array(geom_offsets, arrow_coords).cast(
            multipoint_storage_type(
                interleaved=True,
                dims=dims,
                large_list=is_large_offset,
            ),
        )
        extension_metadata["ARROW:extension:name"] = "geoarrow.multipoint"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    if geom_type == GeometryType.MULTILINESTRING:
        assert len(offsets) == 2, "Expected two offsets arrays"
        ring_offsets, geom_offsets = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_rings = list_array(ring_offsets, arrow_coords)
        arrow_geoms = list_array(geom_offsets, arrow_rings).cast(
            multilinestring_storage_type(
                interleaved=True,
                dims=dims,
                large_list=is_large_offset,
            ),
        )
        extension_metadata["ARROW:extension:name"] = "geoarrow.multilinestring"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    if geom_type == GeometryType.MULTIPOLYGON:
        assert len(offsets) == 3, "Expected three offsets arrays"
        ring_offsets, polygon_offsets, geom_offsets = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_rings = list_array(ring_offsets, arrow_coords)
        arrow_polygons = list_array(polygon_offsets, arrow_rings)
        arrow_geoms = list_array(geom_offsets, arrow_polygons).cast(
            multipolygon_storage_type(
                interleaved=True,
                dims=dims,
                large_list=is_large_offset,
            ),
        )
        extension_metadata["ARROW:extension:name"] = "geoarrow.multipolygon"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    raise ValueError(f"Unsupported type for geoarrow: {geom_type}")
